/* cpuctl revision history */

/*
1.00 Initial release, fixed bug with near conditional jmp at 088-286 CPU
1.10 More exact error mesages at 16-bit CPUs
1.20 New #GPF handler, for preventing hanging during operating with missed MSRs
1.30 Detection #GPF source was added 
1.40 MSR DUMP was added, bug fix 
1.50 experimental variant for test Ctrl+C handler 
1.55R Release with Ctrl+C handler;
	changes in the program exiting code, minor changes
1.60R goto operators were deleted 
1.70  C6.00/MASM6.00 compiler, bug fix with init values and val64 str
1.75  minor changes, comments have been changed
1.80R add checking all value strings for un-digital (and unmask) symbols
1.90R C5.10/MASM5.1 again, minor changes
2.00R custom Ctrl+C handler currently for preventing #GPF vector damaging only,
        MSR LIST output can be break just by pressing any key on keyboard
*/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <dos.h>
#include <errno.h>

/* misc defines */
typedef	unsigned long	ulong;
typedef unsigned int	uint;
typedef unsigned short	ushort;
typedef unsigned char	uchar;

#define MAXLEN    16
#define READ  	0x01
#define LIST  	0x02
#define WRITE 	0x04

#define OR   	0x10
#define AND  	0x20
#define XOR  	0x40
#define MASK	0x80

#define GPF     0x0D  /* address of GPF#   vector*/
#define CNTC	0x23  /* Ctlr+Break/C exit vector*/

#define DEBUG1	0	/* Debug int vectors */
#define DEBUG2	0	/* Debug values for write */

/* special error codes */
#define		NOERROR		0x00
#define		ERRORDOSVER	0x01
#define		I8086		0x02
#define		I286		0x03
#define		NOCPUID		0x04
#define	 	NOMSR		0x05

/* global vars */
uchar op = 0;
uchar dis_ic = 0;

void far* gpf_vect  = NULL;  /* original #GPF vector */
void far* cntc_vect = NULL;  /* original Cntrl+C vector */

//uchar cntc_flag = 0;    /* flag, that Ctrl+C was pressed */
uint error = EZERO;	/* global error variable set to "no error" */		

/* static vars */
static char h32_str[] = "________";
static char l32_str[] = "________";

        /* vars for write mask (*) function */
static char str_write[] = "0000000000000000";
static char hstr_read[] = "0000000000000000";
static char lstr_read[] = "00000000";

/* extern functions */
extern
int cpuchk(void);

/*
 read MSR reg32, return values from *h32 and *l32 pointerns
*/
extern
int read_msr(ulong reg32, ulong *h32, ulong *l32);

/* write to MSR reg32 h32 and l32, return 0
   function used global op flag
*/
extern
int write_msr(ulong reg32, ulong *h32, ulong *l32);

/* new #GPF vector */
extern
interrupt far new_gpf_vector(void);

/* new Cntrl+C vector */
extern
interrupt far new_cntc_vector(void);

/*Convert ASCII string to two 32-bit val h32 and l32 */
int get_h_l_val(char *valstr, ulong *h32, ulong *l32) {
    uint len = 0;
    uchar i;

    len = strlen(valstr);
    if (len <= 8) {
	sscanf(valstr, "%lx", l32);
	*h32 = 0;
    } else {
        strncpy(h32_str, valstr, len - 8);
	sscanf(h32_str, "%lx", h32);

        for (i = 0; i <= 7; i++)
	l32_str[i] = valstr[i + len - 8];
        //l32_str[i] = 0; /* l32_str already has ending zero */
	sscanf(l32_str, "%lx", l32);
    }
return EZERO;
}

void usage(void) {
    printf("CPUCTL v2.00R, (c) 2016 - 2023 Emil Mammadov\n");
    printf("CPUCTL [R W WD] reg32 [reg32end | val64] R - read MSR; W - write MSR\n");
    printf("R reg32          - read one MSR\n");
    printf("R reg32 reg32end - show list of MSR registers from reg32 to reg32end\n");
    printf("W     reg32 - write MSR reg32\n");
    printf("WD    reg32 - disable interrupts and caches during write to MSR reg32\n");
    printf("W[D]A reg32 - (val64 & MSR value) and write MSR reg32\n");
    printf("W[D]O reg32 - (val64 | MSR value) and write MSR reg32\n");
    printf("W[D]X reg32 - (val64 ^ MSR value) and write MSR reg32\n");
    printf("W[D]M reg32 - (val64 mask) do not touch hex digits that were labeled as '*'\n");
}

/*--------------------------------------------------------------------------*/
int main(int argc, char *argv[]) {
    ulong reg32 = 0;
    ulong reg32end = 0;
    ulong h32r, l32r;
    ulong h32w = 0;
    ulong l32w = 0;
    uint ret = 0;
    uint len = 0;
    uint i;

#if DEBUG1
    void far* new_gpf_dbg  = NULL;
    void far* new_cntc_dbg = NULL;
#endif

    if (argc <= 2) {
	usage();
	return (EINVAL);
    }

    ret = cpuchk();

    switch (ret) {
	case NOERROR:
	    break;
	case ERRORDOSVER:
	    printf("DOS version less than 2.x was found, exit\n");
	    return (ENXIO);
	    break;
	case I8086:
	    printf("8086/88 16-bit CPU was found, exit\n");
	    return (ENXIO);
	    break;
	case I286:
	    printf("286 16-bit CPU was found, exit\n");
	    return (ENXIO);
	    break;
	case NOCPUID:
	    printf("CPU doesn't CPUID support, exit\n");
	    return (ENXIO);
	    break;
	case NOMSR:
	    printf("CPU doesn't MSR support, exit\n");
	    return (ENXIO);
	    break;
	default:
	    printf("Unknown error, exit\n");
	    return (ENXIO);
	    break;
    }

    strupr(argv[1]);	/* convert to uppercase CMD keys str */

    if (strncmp(argv[1], "R", 1) == 0 && argc == 3 &&
                strlen(argv[2]) <= 8) 
	op = READ;
    else if (strncmp(argv[1], "R", 1) == 0 && argc == 4 &&
                strlen(argv[2]) <= 8 && strlen(argv[3]) <= 8) 
        op = LIST;
    else if (strncmp(argv[1], "W",1) == 0 && argc == 4 &&
                strlen(argv[2]) <= 8 && strlen(argv[3]) <= 16)
	op = WRITE;
    else {
	usage();
	return (EINVAL);
    }

    /* set flags */
    if (op == WRITE) {
        if (strchr(argv[1], 'D') != NULL)
            dis_ic = 1; /* set "disable int and cache" flag */
        if (strchr(argv[1], 'A'))
            op |= AND;
        else if (strchr(argv[1], 'O'))
            op |= OR;
        else if (strchr(argv[1], 'X'))
            op |= XOR;
        else if (strchr(argv[1], 'M'))
            op |= MASK;
    }

/*---------------------------------------------------------------------------*/
/* Check all symbols in the values's strings, if any symbol
	is not true HEX value, then EINVAL code will be returned*/
    strupr(argv[2]);    /* convert to uppercase MSR reg string */
    len = strlen(argv[2]);
    for (i = 0; i < len; i++) {
	if ( (argv[2][i] < '0') || (argv[2][i] > 'F') ||
		((argv[2][i] > '9')  && (argv[2][i] < 'A')) ) {
		usage();
		return (EINVAL);
	}
    }

/* convert to uppercase new MSR value string or MSR reg32end string */
    strupr(argv[3]);
    len = strlen(argv[3]);
/*   for WM operation, we additionally allow the '*' symbol in value string */
    if ( op == (WRITE | MASK)) {
	for (i=0; i < len; i++) {
	    if (  ((argv[3][i] < '0') || (argv[3][i] > 'F') ||
			((argv[3][i] > '9')  && (argv[3][i] < 'A')))  &&
			(argv[3][i] != '*')  ) {
		usage();
		return (EINVAL);
	   }
	}
    } else {
	for (i=0; i < len; i++) {
	    if ( (argv[3][i] < '0') || (argv[3][i] > 'F') ||
		((argv[3][i] > '9')  && (argv[3][i] < 'A')) ) {
		usage();
		return (EINVAL);
	    }
	}
    }

/*---------------------------------------------------------------------------*/
/* save and change int 0x0D (#GPF) vector
    for prevent halting, if MSR doesn't exist */
    gpf_vect =(void far*) _dos_getvect(GPF);
    _dos_setvect(GPF, new_gpf_vector); 

/* save and change int 0x23 (Cntrl+C) vector for
    for prevent loss the original #GPF vector, if Cntrl+C event will happen */
    cntc_vect = (void far*) _dos_getvect(CNTC);
    _dos_setvect(CNTC, new_cntc_vector); 

#if DEBUG1
    printf("Original #GPT vector   : %04X:%04X\n",
		FP_SEG(gpf_vect), FP_OFF(gpf_vect));
    printf("Original Cntrl+C vector: %04X:%04X\n",
		FP_SEG(cntc_vect), FP_OFF(cntc_vect));

    new_gpf_dbg = (void far*) _dos_getvect(GPF);
    printf("New #GPT vector        : %04X:%04X\n",
                FP_SEG(new_gpf_dbg), FP_OFF(new_gpf_dbg));

    new_cntc_dbg =(void far*) _dos_getvect(CNTC);
    printf("New Cntrl+C vector     : %04X:%04X\n",
        FP_SEG(new_cntc_dbg), FP_OFF(new_cntc_dbg));
#endif

    switch (op) {
        case READ:      /* show one MSR */
	    sscanf(argv[2], "%lx", &reg32);
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		    printf("REG %08lX =", reg32);
		    printf(" %08lX",  h32r);
		    printf(" %08lX\n", l32r);
		}
	    break;

        case LIST:	/* show MSR list */
	    sscanf(argv[2], "%lx", &reg32);
	    sscanf(argv[3], "%lx", &reg32end);
	    for (reg32; reg32 <= reg32end; reg32++) {
		/* check for any key was pressed */
		if (kbhit() != 0)
		    break;
		if (read_msr(reg32, &h32r, &l32r) != NOMSR) {
		    printf("REG %08lX =", reg32);
		    printf(" %08lX",  h32r);
		    printf(" %08lX\n", l32r);
		} 
	    } /* end of for cycle */
	    break;

	case WRITE:
	    sscanf(argv[2], "%lx", &reg32);
            /* read just for test, that MSR is existed */
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		get_h_l_val(argv[3], &h32w, &l32w);
#if DEBUG2
		printf("val for write %08lX %08lX\n", h32w, l32w);
#endif
		write_msr(reg32, &h32w, &l32w);
	    }
	    break;

	case (WRITE | AND):
            sscanf(argv[2], "%lx", &reg32);
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		get_h_l_val(argv[3], &h32w, &l32w);
		h32w &= h32r;
		l32w &= l32r;
#if DEBUG2
		printf("val for write %08lX %08lX\n", h32w, l32w);
#endif
		write_msr(reg32, &h32w, &l32w);
	    }
	    break;

	case (WRITE | OR):
            sscanf(argv[2], "%lx", &reg32);
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		get_h_l_val(argv[3], &h32w, &l32w);
		h32w |= h32r;
		l32w |= l32r;
#if DEBUG2
		printf("val for write %08lX %08lX\n", h32w, l32w);
#endif
		write_msr(reg32, &h32w, &l32w);
	    }
	    break;

	case (WRITE | XOR):
            sscanf(argv[2], "%lx", &reg32);
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		get_h_l_val(argv[3], &h32w, &l32w);
		h32w ^= h32r;
		l32w ^= l32r;
#if DEBUG2
		printf("val for write %08lX %08lX\n", h32w, l32w);
#endif
		write_msr(reg32, &h32w, &l32w);
	    }
	    break;

	case (WRITE | MASK):
            sscanf(argv[2], "%lx", &reg32);
	    if (read_msr(reg32, &h32r, &l32r) == NOMSR)
		error = ENXIO;
	    else {
		sprintf(hstr_read, "%08lx", h32r);
		sprintf(lstr_read, "%08lx", l32r);
		strcat(hstr_read, lstr_read); /* here we get string from MSR reg */

		len = strlen(argv[3]);
		/* here we get mask str with zero at the begin of the string */
		for (i=0; i < len; i++)
		    str_write[MAXLEN - 1 - i] = argv[3][len - 1 - i];

		 /* here we get string for write */
		for (i=0; i < MAXLEN; i++)
		    if (str_write[i] == '*')
			str_write[i] = hstr_read[i];
		get_h_l_val(str_write, &h32w, &l32w);
#if DEBUG2
		printf("val for write %08lX %08lX\n", h32w, l32w);
#endif
		write_msr(reg32, &h32w, &l32w);
	    }
	    break;
	/* unreacheable */
	default:
	    usage();
	    error = EINVAL;	   
    }

    if (error == ENXIO)	 
	printf("Requested MSR is not existed!\n");

    /* restore original #GPF vector */
    _dos_setvect(GPF, (void far*) gpf_vect);

    /* restore original Cntrl+C vector */
    _dos_setvect(CNTC, (void far*) cntc_vect); 

#if DEBUG1
    new_gpf_dbg = (void far*)_dos_getvect(GPF);
    printf("Restored #GPT vector   : %04X:%04X\n",
            FP_SEG(new_gpf_dbg), FP_OFF(new_gpf_dbg));

    new_cntc_dbg = (void far*) _dos_getvect(CNTC);
    printf("Restored Cntrl+C vector: %04X:%04X\n",
        FP_SEG(new_cntc_dbg), FP_OFF(new_cntc_dbg));
#endif
    exit (error);
}
