        INCLUDE constant.inc
        INCLUDE instr.inc
        PUBLIC _cpuchk, _read_msr, _write_msr
        PUBLIC _new_gpf_vector, _new_cntc_vector
        EXTRN _dis_ic:byte
        EXTRN _gpf_vect:dword    ; original GPF# vector

.386P
_TEXT SEGMENT WORD PUBLIC USE16 'CODE'
ASSUME CS:_TEXT
_cpuchk PROC NEAR
.8086	; conditional jumps at a 16-bit CPU will be short
    push bx
    push cx
    push dx
    push si
    push di
    pushf

; checking DOS version
chk_dos:
    mov ah, 30h         ; get DOS version in AL
    int 21h
    cmp al, 02
    jae chk_32bit       ; if DOS version above or equal 3.00, next check
    mov ax, ERRORDOSVER
    jmp short exit_err

; does CPU 32 bit support?
chk_32bit:
                        ; check 8086, try to clear bits 15-12
    pushf
    pop ax
    and ax, 0fffh	; clear bits 15-12 in AX
    push ax
    popf
    pushf
    pop ax
    and ax, 0f000h	; & mask bits 15-12
    cmp ax, 0f000h	; if bits 15 - 12 are equal 1111, 8086 is detected
    je i8086_found		; 8086 is detected

; check 80286, try to set bits 15-12
    pushf
    pop ax
    or ax, 0f000h	; set bits 15-12 in AX
    push ax
    popf
    pushf
    pop ax
    and ax, 0f000h	; & mask bits 15-12
    cmp ax, 0		; if bits 15 - 12 are equal 0000, 80286 is detected
    je i286_found		; 80286 is detected

    jmp short chk_cpuid       ; if CPU(8086/80286 were not detected), next check
i8086_found:
    mov ax, I8086
    jmp short exit_err
i286_found:
    mov ax, I286
    jmp short exit_err

.386P
; does CPU does CPUID support?
chk_cpuid:
; bit 21 EFLAGS possible to change if CPUID instruction is supported
    pushfd
    pop eax
    mov ebx, eax
    and ebx, BIT21 ; & mask bit 21
    xor eax, BIT21 ; ~bit 21
    push eax
    popfd

    pushfd
    pop eax
    and eax, BIT21
    cmp eax, ebx ;if inversion possible, CPUID is supported
    jne short chk_msr  ;if our CPU does CPUID support, next check
    mov ax, NOCPUID
    jmp short exit_err

; does CPU does MSR support?
chk_msr:
    mov eax, 01h
    cpuid
    test edx, 020h ;test bit 5(MSR support)
    jnz short exit_no_error ;if our CPU does MSR support, then exit with NOERROR code
    mov ax, NOMSR ;error code - no MSR support
    jmp  short exit_err

exit_no_error:
    mov ax, NOERROR
exit_err:
    popf
    pop di
    pop si
    pop dx
    pop cx
    pop bx
    retn
_cpuchk ENDP

;---------------------------------------------------------------------------
_write_msr PROC NEAR
    ; reg32 [bp+4]
    ; &h32  [bp+8]
    ; &l32  [bp+10]
    enter 0, 0  ;Entry sequence

    push bx
    push cx
    push dx
    push si
    push di

    test  _dis_ic, 01
    je short sdc
    cli         ; disable interrupts
    mov eax, cr0
    or eax, 60000000h ;set bits CD(30) and NW(29)
    wbinvd ;clearing L1/L2 caches
    mov cr0, eax ; disable caching
sdc:

    ; get reg32 in ECX
    mov cx, [bp+6]	;get high word
    shl ecx, 16
    mov cx, [bp+4]	;get low word

    mov si, [bp+8]	;get &h32
    mov dx, [si+2]
    ror edx, 16
    mov dx, [si]

    mov si, [bp+10]	; get &l32
    mov ax, [si+2]
    ror eax, 16
    mov ax, [si]

    wrmsr

    test _dis_ic, 01
    je short sec
    mov eax, cr0
    and eax, 09fffffffh	; clear bits CD(30) and NW(29)
    mov cr0, eax	; enable caching(really only L1)
    sti			; enable interrupts
sec:
    pop di
    pop si
    pop dx
    pop cx
    pop bx

    leave       ; Exit sequence
    mov ax, 0

    retn
_write_msr ENDP

;---------------------------------------------------------------------------
_read_msr PROC NEAR
    ; reg32 [bp+4]
    ; &h32  [bp+8]
    ; &l32  [bp+10]
    
    enter 0, 0  ;Entry sequence

    push bx
    push cx
    push dx
    push si
    push di

    ;get reg32 in ECX
    mov cx, [bp+6]	;get high word
    shl ecx, 16
    mov cx, [bp+4]	;get low word

    ;pushf
    ;call far ptr _new_gpf_vector

read_msr_addr:
    rdmsr

    mov si, [bp+8]	;get &h32
    mov [si], dx
    ror edx, 16
    mov [si+2], dx

    mov si, [bp+10]     ; get &l32
    mov [si], ax
    ror eax, 16
    mov [si+2], ax

    mov ax, 0
    jmp short r1         
err_no_msr:
    mov ax, NOMSR      ; jmp from new GPF# handler, if we have no this MSR

r1:
    pop di
    pop si
    pop dx
    pop cx
    pop bx

    leave              ;Exit sequence

    retn
_read_msr ENDP

;---------------------------------------------------------------------------
_new_gpf_vector PROC FAR
    enter    0, 0
    pusha

    ; if we were called from another source (not from CPUCTL.EXE)
    ; then we will call old #GPF handler
    mov bx, [bp+2] ; real offset
    mov cx, [bp+4] ; real segment

   ; compare real offset (was saved in stack) and rdmsr segment
    cmp bx, offset read_msr_addr
    jne short jmp_to_org_gpf

    mov dx, cs
    ; compare real segment (was saved in stack) and our proc segment
    cmp cx, dx   
    jne short jmp_to_org_gpf    
            ;jmp short  jmp_to_org_gpf   ;for debug only!!! 

    lea ax, cs:err_no_msr
    mov     [bp+2], ax
    ;mov     [bp+4], cs    ; it isn't needed, since all code in same segment

    popa
    leave
    iret

jmp_to_org_gpf:
    popa
    leave
    jmp dword ptr _gpf_vect
_new_gpf_vector ENDP

;---------------------------------------------------------------------------
_new_cntc_vector PROC FAR
; just do nothing
    iret
_new_cntc_vector ENDP

_TEXT ENDS
     END

